#
# API Gateway: Cliente de API en Python.
# Copyright (C) API Gateway <https://www.apigateway.cl>
#
# Este programa es software libre: usted puede redistribuirlo y/o modificarlo
# bajo los términos de la GNU Lesser General Public License (LGPL) publicada
# por la Fundación para el Software Libre, ya sea la versión 3 de la Licencia,
# o (a su elección) cualquier versión posterior de la misma.
#
# Este programa se distribuye con la esperanza de que sea útil, pero SIN
# GARANTÍA ALGUNA; ni siquiera la garantía implícita MERCANTIL o de APTITUD
# PARA UN PROPÓSITO DETERMINADO. Consulte los detalles de la GNU Lesser General
# Public License (LGPL) para obtener una información más detallada.
#
# Debería haber recibido una copia de la GNU Lesser General Public License
# (LGPL) junto a este programa. En caso contrario, consulte
# <http://www.gnu.org/licenses/lgpl.html>.
#



#
# API Gateway: Cliente de API en Python.
# Copyright (C) API Gateway <https://www.apigateway.cl>
#
# Este programa es software libre: usted puede redistribuirlo y/o modificarlo
# bajo los términos de la GNU Lesser General Public License (LGPL) publicada
# por la Fundación para el Software Libre, ya sea la versión 3 de la Licencia,
# o (a su elección) cualquier versión posterior de la misma.
#
# Este programa se distribuye con la esperanza de que sea útil, pero SIN
# GARANTÍA ALGUNA; ni siquiera la garantía implícita MERCANTIL o de APTITUD
# PARA UN PROPÓSITO DETERMINADO. Consulte los detalles de la GNU Lesser General
# Public License (LGPL) para obtener una información más detallada.
#
# Debería haber recibido una copia de la GNU Lesser General Public License
# (LGPL) junto a este programa. En caso contrario, consulte
# <http://www.gnu.org/licenses/lgpl.html>.
#

'''
Módulo para obtener las actividades económicas del SII.

Para más información sobre la API, consulte la `documentación completa de
Actividades Económicas <https://developers.apigateway.cl/#e64eb128-173a-48c7-ab0b-b6152e59c327>`_.
'''

from .. import ApiBase

class ActividadesEconomicas(ApiBase):
    '''
    Cliente específico para interactuar con los endpoints de actividades
    económicas de la API de API Gateway.

    Provee métodos para obtener listados de actividades económicas, tanto
    de primera como de segunda categoría.
    '''

    def listado(self, categoria = None):
        '''
        Obtiene un listado de actividades económicas. Puede filtrar por categoría.

        :param int categoria: Categoría de las actividades económicas (opcional).
        :return: Respuesta JSON con el listado de actividades económicas.
        :rtype: list[dict]
        '''
        url = '/sii/contribuyentes/actividades_economicas'
        if categoria is not None:
            url += '/%(categoria)s' % {'categoria': categoria}
        response = self.client.get(url)
        return response.json()

    def listado_primera_categoria(self):
        '''
        Obtiene un listado de actividades económicas de primera categoría.

        :return: Respuesta JSON con el listado de actividades económicas de primera categoría.
        :rtype: list[dict]
        '''
        return self.listado(1)

    def listado_segunda_categoria(self):
        '''
        Obtiene un listado de actividades económicas de segunda categoría.

        :return: Respuesta JSON con el listado de actividades económicas de segunda categoría.
        :rtype: list[dict]
        '''
        return self.listado(2)



#
# API Gateway: Cliente de API en Python.
# Copyright (C) API Gateway <https://www.apigateway.cl>
#
# Este programa es software libre: usted puede redistribuirlo y/o modificarlo
# bajo los términos de la GNU Lesser General Public License (LGPL) publicada
# por la Fundación para el Software Libre, ya sea la versión 3 de la Licencia,
# o (a su elección) cualquier versión posterior de la misma.
#
# Este programa se distribuye con la esperanza de que sea útil, pero SIN
# GARANTÍA ALGUNA; ni siquiera la garantía implícita MERCANTIL o de APTITUD
# PARA UN PROPÓSITO DETERMINADO. Consulte los detalles de la GNU Lesser General
# Public License (LGPL) para obtener una información más detallada.
#
# Debería haber recibido una copia de la GNU Lesser General Public License
# (LGPL) junto a este programa. En caso contrario, consulte
# <http://www.gnu.org/licenses/lgpl.html>.
#

'''
Módulo para interactuar con Boletas de Honorarios Electrónicas, tanto
emitidas como recibidas, del SII.

Para más información sobre la API, consulte la `documentación completa de las
BHE <https://developers.apigateway.cl/#4df9775f-2cd3-4b35-80a5-373f2501230c>`_.
'''

from .. import ApiBase

class BheEmitidas(ApiBase):
    '''
    Cliente específico para gestionar Boletas de Honorarios Electrónicas (BHE) emitidas.

    Provee métodos para emitir, anular, y consultar información relacionada con BHEs.

    :param str identificador: Identificador del contribuyente.
    :param str clave: Clave del identificador.
    :param kwargs: Argumentos adicionales.
    '''

    # Quién debe hacer la retención asociada al honorario para pagar al SII
    RETENCION_RECEPTOR = 1
    RETENCION_EMISOR = 2

    # Posibles motivos de anulación de una BHE
    ANULACION_CAUSA_SIN_PAGO = 1
    ANULACION_CAUSA_SIN_PRESTACION = 2
    ANULACION_CAUSA_ERROR_DIGITACION = 3

    def __init__(self, identificador, clave, **kwargs):
        super().__init__(identificador = identificador, clave = clave, **kwargs)

    def documentos(self, emisor, periodo, pagina = None, pagina_sig_codigo = None):
        '''
        Obtiene los documentos de BHE emitidos por un emisor en un periodo específico.

        :param str emisor: RUT del emisor de las boletas.
        :param str periodo: Período de tiempo de las boletas emitidas.
        :return: Respuesta JSON con los documentos de BHE.
        :rtype: list[dict]
        '''
        url = '/sii/bhe/emitidas/documentos/%(emisor)s/%(periodo)s' % {'emisor': emisor, 'periodo': periodo}
        if pagina is not None:
            url += '?pagina=%(pagina)s' % {
                'pagina': pagina,
            }
            if pagina_sig_codigo:
                url += '&pagina_sig_codigo=%(pagina_sig_codigo)s' % {
                    'pagina_sig_codigo': pagina_sig_codigo or '0'
                }
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post(url, data = body)
        return response.json()

    def emitir(self, boleta):
        '''
        Emite una nueva Boleta de Honorarios Electrónica.

        :param dict boleta: Información detallada de la boleta a emitir.
        :return: Respuesta JSON con la confirmación de la emisión de la BHE.
        :rtype: dict
        '''
        body = {
            'auth': self._get_auth_pass(),
            'boleta': boleta
        }
        response = self.client.post('/sii/bhe/emitidas/emitir', data = body)
        return response.json()

    def pdf(self, codigo):
        '''
        Obtiene el PDF de una BHE emitida.

        :param str codigo: Código único de la BHE.
        :return: Contenido del PDF de la BHE.
        :rtype: bytes
        '''
        url = '/sii/bhe/emitidas/pdf/%(codigo)s' % {'codigo': codigo}
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post(url, data = body)
        return response.content

    def email(self, codigo, email):
        '''
        Envía por correo electrónico una BHE emitida.

        :param str codigo: Código único de la BHE a enviar.
        :param str email: Dirección de correo electrónico a la cual enviar la BHE.
        :return: Respuesta JSON con la confirmación del envío del email.
        :rtype: dict
        '''
        url = '/sii/bhe/emitidas/email/%(codigo)s' % {'codigo': codigo}
        body = {
            'auth': self._get_auth_pass(),
            'destinatario': {'email': email}
        }
        response = self.client.post(url, data = body)
        return response.json()

    def anular(self, emisor, folio, causa = ANULACION_CAUSA_ERROR_DIGITACION):
        '''
        Anula una BHE emitida.

        :param str emisor: RUT del emisor de la boleta.
        :param str folio: Número de folio de la boleta.
        :param int causa: Motivo de anulación de la boleta.
        :return: Respuesta JSON con la confirmación de la anulación de la BHE.
        :rtype: dict
        '''
        url = '/sii/bhe/emitidas/anular/%(emisor)s/%(folio)s?causa=%(causa)s' % {
            'emisor': emisor, 'folio': folio, 'causa': causa
        }
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post(url, data = body)
        return response.json()

class BheRecibidas(ApiBase):
    '''
    Cliente específico para gestionar Boletas de Honorarios Electrónicas (BHE) recibidas.

    Provee métodos para obtener documentos, obtener PDF y observar BHE recibidas.

    :param str identificador: Identificador del contribuyente.
    :param str clave: Clave del identificador.
    :param kwargs: Argumentos adicionales.
    '''

    def __init__(self, identificador, clave, **kwargs):
        super().__init__(identificador = identificador, clave = clave, **kwargs)

    def documentos(self, receptor, periodo, pagina = None, pagina_sig_codigo = None):
        '''
        Obtiene los documentos de BHE recibidos por un receptor en un periodo específico.

        :param str receptor: RUT del receptor de las boletas.
        :param str periodo: Período de tiempo de las boletas recibidas.
        :param int pagina: Número de página para paginación (opcional).
        :param str pagina_sig_codigo: Código para la siguiente página (opcional).
        :return: Respuesta JSON con los documentos de BHE.
        :rtype: list[dict]
        '''
        url = '/sii/bhe/recibidas/documentos/%(receptor)s/%(periodo)s' % {
            'receptor': receptor, 'periodo': periodo
        }
        body = {
            'auth': self._get_auth_pass()
        }
        if pagina is not None:
            url += '?pagina=%(pagina)s&pagina_sig_codigo=%(pagina_sig_codigo)s' % {
                'pagina': pagina,
                'pagina_sig_codigo': pagina_sig_codigo or '00000000000000'
            }
        response = self.client.post(url, data = body)
        return response.json()

    def pdf(self, codigo):
        '''
        Obtiene el PDF de una BHE recibida.

        :param str codigo: Código único de la BHE.
        :return: Contenido del PDF de la BHE.
        :rtype: bytes
        '''
        url = '/sii/bhe/recibidas/pdf/%(codigo)s' % {'codigo': codigo}
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post(url, data = body)
        return response.content

    def observar(self, emisor, numero, causa = 1):
        '''
        Marca una observación en una BHE recibida.

        :param str emisor: RUT del emisor de la boleta.
        :param str numero: Número de la boleta.
        :param int causa: Motivo de la observación.
        :return: Respuesta JSON con la confirmación de la observación.
        :rtype: dict
        '''
        url = '/sii/bhe/recibidas/observar/%(emisor)s/%(numero)s?causa=%(causa)s' % {
            'emisor': emisor, 'numero': numero, 'causa': causa
        }
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post(url, data = body)
        return response.json()



#
# API Gateway: Cliente de API en Python.
# Copyright (C) API Gateway <https://www.apigateway.cl>
#
# Este programa es software libre: usted puede redistribuirlo y/o modificarlo
# bajo los términos de la GNU Lesser General Public License (LGPL) publicada
# por la Fundación para el Software Libre, ya sea la versión 3 de la Licencia,
# o (a su elección) cualquier versión posterior de la misma.
#
# Este programa se distribuye con la esperanza de que sea útil, pero SIN
# GARANTÍA ALGUNA; ni siquiera la garantía implícita MERCANTIL o de APTITUD
# PARA UN PROPÓSITO DETERMINADO. Consulte los detalles de la GNU Lesser General
# Public License (LGPL) para obtener una información más detallada.
#
# Debería haber recibido una copia de la GNU Lesser General Public License
# (LGPL) junto a este programa. En caso contrario, consulte
# <http://www.gnu.org/licenses/lgpl.html>.
#

'''
Módulo para la emisión de Boletas de Terceros Electrónicas del SII.

Para más información sobre la API, consulte la `documentación completa de las
BTE <https://developers.apigateway.cl/#e08f50ab-5509-48ab-81ab-63fc8e5985e1>`_.
'''

from .. import ApiBase

class BteEmitidas(ApiBase):
    '''
    Cliente específico para gestionar Boletas de Terceros Electrónicas (BTE) emitidas.

    Provee métodos para emitir, anular, y consultar información relacionada con BTEs.

    :param str identificador: Identificador del contribuyente.
    :param str clave: Clave del identificador.
    :param kwargs: Argumentos adicionales.
    '''

    def __init__(self, identificador, clave, **kwargs):
        super().__init__(identificador = identificador, clave = clave, **kwargs)

    def documentos(self, emisor, periodo, pagina = None):
        '''
        Obtiene los documentos BTE emitidos por un emisor en un periodo específico.

        :param str emisor: RUT del emisor de las BTE.
        :param str periodo: Período de las BTE emitidas.
        :return: Respuesta JSON con los documentos BTE.
        :rtype: list[dict]
        '''
        url = '/sii/bte/emitidas/documentos/%(emisor)s/%(periodo)s' % {
            'emisor': emisor, 'periodo': periodo
        }
        if pagina is not None:
            url += '?pagina=%(pagina)s' % {
                'pagina': pagina,
            }
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post(url, data = body)
        return response.json()

    def html(self, codigo):
        '''
        Obtiene la representación HTML de una BTE emitida.

        :param str codigo: Código único de la BTE.
        :return: Contenido HTML de la BTE.
        :rtype: str
        '''
        url = '/sii/bte/emitidas/html/%(codigo)s' % {'codigo': codigo}
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post(url, data = body)
        return response.content

    def emitir(self, datos):
        '''
        Emite una nueva Boleta de Tercero Electrónica.

        :param dict datos: Datos de la boleta a emitir.
        :return: Respuesta JSON con la confirmación de la emisión de la BTE.
        :rtype: dict
        '''
        body = {
            'auth': self._get_auth_pass(),
            'boleta': datos
        }
        response = self.client.post('/sii/bte/emitidas/emitir', data = body)
        return response.json()

    def anular(self, emisor, numero, causa = 3, periodo = None):
        '''
        Anula una BTE emitida.

        :param str emisor: RUT del emisor de la boleta.
        :param str numero: Número de la boleta.
        :param int causa: Causa de anulación.
        :param str periodo: Período de emisión de la boleta (opcional).
        :return: Respuesta JSON con la confirmación de la anulación.
        :rtype: dict
        '''
        body = {
            'auth': self._get_auth_pass()
        }
        url = '/sii/bte/emitidas/anular/%(emisor)s/%(numero)s?causa=%(causa)s' % {
            'emisor': emisor, 'numero': numero, 'causa': causa
        }
        if periodo:
            url += '&periodo=%(periodo)s' % {'periodo': periodo}
        response = self.client.post(url, data = body)
        return response.json()

    def receptor_tasa(self, emisor, receptor, periodo = None):
        '''
        Obtiene la tasa de retención aplicada a un receptor por un emisor específico.

        :param str emisor: RUT del emisor de la boleta.
        :param str receptor: RUT del receptor de la boleta.
        :param str periodo: Período de emisión de la boleta (opcional).
        :return: Respuesta JSON con la tasa de retención.
        :rtype: dict
        '''
        body = {
            'auth': self._get_auth_pass()
        }
        url = '/sii/bte/emitidas/receptor_tasa/%(emisor)s/%(receptor)s' % {
            'emisor': emisor, 'receptor': receptor
        }
        if periodo:
            url += '?periodo=%(periodo)s' % {'periodo': periodo}
        response = self.client.post(url, data = body)
        return response.json()



#
# API Gateway: Cliente de API en Python.
# Copyright (C) API Gateway <https://www.apigateway.cl>
#
# Este programa es software libre: usted puede redistribuirlo y/o modificarlo
# bajo los términos de la GNU Lesser General Public License (LGPL) publicada
# por la Fundación para el Software Libre, ya sea la versión 3 de la Licencia,
# o (a su elección) cualquier versión posterior de la misma.
#
# Este programa se distribuye con la esperanza de que sea útil, pero SIN
# GARANTÍA ALGUNA; ni siquiera la garantía implícita MERCANTIL o de APTITUD
# PARA UN PROPÓSITO DETERMINADO. Consulte los detalles de la GNU Lesser General
# Public License (LGPL) para obtener una información más detallada.
#
# Debería haber recibido una copia de la GNU Lesser General Public License
# (LGPL) junto a este programa. En caso contrario, consulte
# <http://www.gnu.org/licenses/lgpl.html>.
#

'''
Módulo para interactuar con el Registro de Compra y Venta del SII.

Para más información sobre la API, consulte la `documentación completa del
RCV <https://developers.apigateway.cl/#ef1f7d54-2e86-4732-bb91-d3448b383d66>`_.
'''

from .. import ApiBase

class Rcv(ApiBase):
    '''
    Cliente específico para interactuar con los endpoints de Registro de Compras
    y Ventas (RCV) de la API de API Gateway.

    Proporciona métodos para obtener resúmenes y detalles de compras y ventas.

    :param str identificador: Identificador del contribuyente.
    :param str clave: Clave del identificador.
    :param kwargs: Argumentos adicionales.
    '''

    def __init__(self, identificador, clave, **kwargs):
        super().__init__(identificador = identificador, clave = clave, **kwargs)

    def compras_resumen(self, receptor, periodo, estado = 'REGISTRO'):
        '''
        Obtiene un resumen de las compras registradas para un receptor en un periodo específico.

        :param str receptor: RUT del receptor de las compras.
        :param str periodo: Período de tiempo de las compras.
        :param str estado: Estado de las compras ('REGISTRO', 'PENDIENTE', 'NO_INCLUIR', 'RECLAMADO').
        :return: Respuesta JSON con el resumen de compras.
        :rtype: list[dict]
        '''
        url = '/sii/rcv/compras/resumen/%(receptor)s/%(periodo)s/%(estado)s' % {
            'receptor': receptor, 'periodo': periodo, 'estado': estado
        }
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post( url, data = body)
        return response.json()

    def compras_detalle(self, receptor, periodo, dte = 0, estado = 'REGISTRO', tipo = None):
        '''
        Obtiene detalles de las compras para un receptor en un periodo específico.

        :param str receptor: RUT del receptor de las compras.
        :param str periodo: Período de tiempo de las compras.
        :param int dte: Tipo de DTE.
        :param str estado: Estado de las compras ('REGISTRO', 'PENDIENTE', 'NO_INCLUIR', 'RECLAMADO').
        :param str tipo: Tipo de formato de respuesta ('rcv_csv' o 'rcv').
        :return: Respuesta JSON con detalles de las compras.
        :rtype: list[dict]
        '''
        url = '/sii/rcv/compras/detalle/%(receptor)s/%(periodo)s/%(dte)s/%(estado)s?tipo=%(tipo)s' % {
            'receptor': receptor,
            'periodo': periodo,
            'dte': dte,
            'estado': estado,
            'tipo': tipo
        }
        tipo = 'rcv_csv' if dte == 0 and estado == 'REGISTRO' else tipo or 'rcv'
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post(url, data = body)
        return response.json()

    def ventas_resumen(self, emisor, periodo):
        '''
        Obtiene un resumen de las ventas registradas para un emisor en un periodo específico.

        :param str emisor: RUT del emisor de las ventas.
        :param str periodo: Período de tiempo de las ventas.
        :return: Respuesta JSON con el resumen de ventas.
        :rtype: list[dict]
        '''
        url = '/sii/rcv/ventas/resumen/%(emisor)s/%(periodo)s' % {
            'emisor': emisor, 'periodo': periodo
        }
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post(url, data = body)
        return response.json()

    def ventas_detalle(self, emisor, periodo, dte = 0, tipo = None):
        '''
        Obtiene detalles de las ventas para un emisor en un periodo específico.

        :param str emisor: RUT del emisor de las ventas.
        :param str periodo: Período de tiempo de las ventas.
        :param int dte: Tipo de DTE.
        :param str tipo: Tipo de formato de respuesta ('rcv_csv' o 'rcv').
        :return: Respuesta JSON con detalles de las ventas.
        :rtype: list[dict]
        '''
        tipo = 'rcv_csv' if dte == 0 else tipo or 'rcv'
        url = '/sii/rcv/ventas/detalle/%(emisor)s/%(periodo)s/%(dte)s?tipo=%(tipo)s' % {
            'emisor': emisor, 'periodo': periodo, 'dte': dte, 'tipo': tipo
        }
        body = {
            'auth': self._get_auth_pass()
        }
        response = self.client.post(url, data = body)
        return response.json()



#
# API Gateway: Cliente de API en Python.
# Copyright (C) API Gateway <https://www.apigateway.cl>
#
# Este programa es software libre: usted puede redistribuirlo y/o modificarlo
# bajo los términos de la GNU Lesser General Public License (LGPL) publicada
# por la Fundación para el Software Libre, ya sea la versión 3 de la Licencia,
# o (a su elección) cualquier versión posterior de la misma.
#
# Este programa se distribuye con la esperanza de que sea útil, pero SIN
# GARANTÍA ALGUNA; ni siquiera la garantía implícita MERCANTIL o de APTITUD
# PARA UN PROPÓSITO DETERMINADO. Consulte los detalles de la GNU Lesser General
# Public License (LGPL) para obtener una información más detallada.
#
# Debería haber recibido una copia de la GNU Lesser General Public License
# (LGPL) junto a este programa. En caso contrario, consulte
# <http://www.gnu.org/licenses/lgpl.html>.
#

from requests.exceptions import Timeout, ConnectionError, RequestException, HTTPError
from os import getenv
from abc import ABC

import urllib.parse
import urllib
import requests
import base64
import json
import re

class ApiClient:
    '''
    Cliente para interactuar con la API de API Gateway.

    :param str token: Token de autenticación del usuario. Si no se proporciona,
    se intentará obtener de una variable de entorno.
    :param str url: URL base de la API. Si no se proporciona, se usará una
    URL por defecto.
    :param str version: Versión de la API. Si no se proporciona, se usará
    una versión por defecto.
    :param bool raise_for_status: Si se debe lanzar una excepción automáticamente
    para respuestas de error HTTP. Por defecto es True.
    '''

    __DEFAULT_URL = 'https://legacy.apigateway.cl'
    __DEFAULT_VERSION = 'v1'

    def __init__(self, token = None, url = None, version = None, raise_for_status = True):
        self.token = self.__validate_token(token)
        self.url = self.__validate_url(url)
        self.headers = self.__generate_headers()
        self.version = version or self.__DEFAULT_VERSION
        self.raise_for_status = raise_for_status

    def __validate_token(self, token):
        '''
        Valida y retorna el token de autenticación.

        :param str token: Token de autenticación a validar.
        :return: Token validado.
        :rtype: str
        :raises ApiException: Si el token no es válido o está ausente.
        '''
        token = token or getenv('APIGATEWAY_API_TOKEN')
        if not token:
            raise ApiException('Se debe configurar la variable de entorno: APIGATEWAY_API_TOKEN.')
        return str(token).strip()

    def __validate_url(self, url):
        '''
        Valida y retorna la URL base para la API.

        :param str url: URL a validar.
        :return: URL validada.
        :rtype: str
        :raises ApiException: Si la URL no es válida o está ausente.
        '''
        return str(url).strip() if url else getenv(
            'APIGATEWAY_API_URL',
            self.__DEFAULT_URL).strip()

    def __generate_headers(self):
        '''
        Genera y retorna las cabeceras por defecto para las solicitudes.

        :return: Cabeceras por defecto.
        :rtype: dict
        '''
        return {
            'User-Agent': 'API Gateway: Cliente de API en Python.',
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': 'Bearer %(token)s' % {'token': self.token}
        }

    def __request(self, method, resource, data = None, headers = None):
        '''
        Método privado para realizar solicitudes HTTP.

        :param str method: Método HTTP a utilizar.
        :param str resource: Recurso de la API a solicitar.
        :param dict data: Datos a enviar en la solicitud (opcional).
        :param dict headers: Cabeceras adicionales para la solicitud (opcional).
        :return: Respuesta de la solicitud.
        :rtype: requests.Response
        :raises ApiException: Si el método HTTP no es soportado o si hay
        un error de conexión.
        '''
        api_path = '/api/%(version)s%(resource)s' % {
            'version': self.version, 'resource': resource
        }
        full_url = urllib.parse.urljoin(self.url + '/', api_path.lstrip('/'))
        headers = headers or {}
        headers = {**self.headers, **headers}
        if data and not isinstance(data, str):
            data = json.dumps(data)
        try:
            response = requests.request(method, full_url, data = data, headers = headers)
            return self.__check_and_return_response(response)
        except ConnectionError as error:
            raise ApiException('Error de conexión: %(error)s' % {
                'error': error
            })
        except Timeout as error:
            raise ApiException('Error de timeout: %(error)s' % {
                'error': error
            })
        except RequestException as error:
            raise ApiException('Error en la solicitud: %(error)s' % {
                'error': error
            })

    def __check_and_return_response(self, response):
        '''
        Verifica la respuesta de la solicitud HTTP y maneja los errores.

        :param requests.Response response: Objeto de respuesta de requests.
        :return: Respuesta validada.
        :rtype: requests.Response
        :raises ApiException: Si la respuesta contiene un error HTTP.
        '''
        if response.status_code != 200 and self.raise_for_status:
            try:
                response.raise_for_status()
            except HTTPError as error:
                try:
                    error = response.json()
                    message = error.get(
                        'message', ''
                    ) or error.get(
                        'exception', ''
                    ) or 'Error desconocido.'
                except json.decoder.JSONDecodeError:
                    message = 'Error al decodificar los datos en JSON: %(response)s' % {
                        'response': response.text
                    }
                raise ApiException('Error HTTP: %(message)s' % {
                    'message': message
                })
        return response

    def get(self, resource, headers = None):
        '''
        Realiza una solicitud GET a la API.

        :param str resource: Recurso de la API a solicitar.
        :param dict headers: Cabeceras adicionales para la solicitud.
        :return: Respuesta de la solicitud.
        :rtype: requests.Response
        '''
        return self.__request('GET', resource, headers = headers)

    def delete(self, resource, headers = None):
        '''
        Realiza una solicitud DELETE a la API.

        :param str resource: Recurso de la API a solicitar.
        :param dict headers: Cabeceras adicionales para la solicitud.
        :return: Respuesta de la solicitud.
        :rtype: requests.Response
        '''
        return self.__request('DELETE', resource, headers = headers)

    def post(self, resource, data = None, headers = None):
        '''
        Realiza una solicitud POST a la API.

        :param str resource: Recurso de la API a solicitar.
        :param dict data: Datos a enviar en la solicitud.
        :param dict headers: Cabeceras adicionales para la solicitud.
        :return: Respuesta de la solicitud.
        :rtype: requests.Response
        '''
        return self.__request('POST', resource, data, headers)

    def put(self, resource, data = None, headers = None):
        '''
        Realiza una solicitud PUT a la API.

        :param str resource: Recurso de la API a solicitar.
        :param dict data: Datos a enviar en la solicitud.
        :param dict headers: Cabeceras adicionales para la solicitud.
        :return: Respuesta de la solicitud.
        :rtype: requests.Response
        '''
        return self.__request('PUT', resource, data, headers)

class ApiException(Exception):
    '''
    Excepción personalizada para errores en el cliente de la API.

    :param str message: Mensaje de error.
    :param int code: Código de error (opcional).
    :param dict params: Parámetros adicionales del error (opcional).
    '''

    def __init__(self, message, code = None, params = None):
        self.message = message
        self.code = code
        self.params = params
        super().__init__(message)

    def __str__(self):
        '''
        Devuelve una representación en cadena del error, proporcionando un contexto claro
        del problema ocurrido. Esta representación incluye el prefijo "[API Gateway]",
        seguido del código de error si está presente, y el mensaje de error.

        Si se especifica un código de error, el formato será:
        "[API Gateway] Error {code}: {message}"

        Si no se especifica un código de error, el formato será:
        "[API Gateway] {message}"

        :return: Una cadena que representa el error de una manera clara y concisa.
        '''
        if self.code is not None:
            return "[API Gateway] Error %(code)s: %(message)s" % {
                'code': self.code,
                'message': self.message
            }
        else:
            return "[API Gateway] %(message)s" % {
                'message': self.message
            }

class ApiBase(ABC):
    '''
    Clase base para las clases que consumen la API (wrappers).

    :param str api_token: Token de autenticación para la API.
    :param str api_url: URL base para la API.
    :param str api_version: Versión de la API.
    :param bool api_raise_for_status: Si se debe lanzar una excepción automáticamente para respuestas de error HTTP. Por defecto es True.
    :param dict kwargs: Argumentos adicionales para la autenticación.
    '''

    auth = {}

    def __init__(
            self,
            api_token = None,
            api_url = None,
            api_version = None,
            api_raise_for_status = True,
            **kwargs
        ):
        self.client = ApiClient(
            api_token,
            api_url,
            api_version,
            api_raise_for_status
        )
        self.__setup_auth(kwargs)

    def __setup_auth(self, kwargs):
        '''
        Configura la autenticación específica para la aplicación.

        :param dict kwargs: Argumentos clave-valor para configurar la autenticación.
        '''
        identificador = kwargs.get('identificador')
        clave = kwargs.get('clave')
        if identificador and clave:
            if self.__is_auth_pass(identificador):
                self.auth = {
                    'pass': {
                        'rut': identificador,
                        'clave': clave
                    }
                }
            elif self.__is_auth_cert_data(identificador):
                self.auth = {
                    'cert': {
                        'cert-data': identificador,
                        'pkey-data': clave
                    }
                }
            elif self.__is_auth_file_data(identificador):
                self.auth = {
                    'cert': {
                        'file-data': identificador,
                        'file-pass': clave
                    }
                }
            else:
                raise ApiException('No se han proporcionado las credenciales de autentificación.')

    def __is_auth_pass(self, rut):
        """
        Valida la estructura de un RUT chileno utilizando una expresión regular.

        Este método verifica que el RUT cumpla con el formato estándar chileno, que incluye
        puntos como separadores de miles opcionales y un guion antes del dígito verificador.
        El dígito verificador puede ser un número o la letra 'K'.

        **Ejemplos de RUT válidos:**
            - 12.345.678-5
            - 12345678-5
            - 9.876.543-K
            - 9876543-K

        **Ejemplos de RUT inválidos:**
            - 12.345.678-9 (dígito verificador incorrecto)
            - 12345678- (falta dígito verificador)
            - 12345-6 (formato incorrecto)
            - 12.345.6785 (falta guion)
            - abcdefgh-i (caracteres no permitidos)

        :param str rut: El RUT a validar.
        :return: True si el RUT tiene un formato válido, False en caso contrario.
        :rtype: bool
        """
        if rut is None:
            return False
        # Expresión regular para validar el formato del RUT chileno
        patron = re.compile(r'^(\d{1,3}\.?)(\d{3}\.?)(\d{3,4})-([\dkK])$')
        return bool(patron.match(rut))

    def __is_auth_file_data(self, firma_electronica_base64):
        """
        Verifica si una cadena es una cadena codificada en Base64 válida.

        :param str firma_electronica_base64: La cadena a verificar.
        :return: True si la cadena es válida en Base64, False en caso contrario.
        :rtype: bool
        """
        if firma_electronica_base64 is None:
            return False
        try:
            # Asegúrate de que la longitud de la cadena sea múltiplo de 4
            if len(firma_electronica_base64) % 4 != 0:
                return False
            # Intenta decodificar la cadena con validación estricta
            base64.b64decode(firma_electronica_base64, validate=True)
            return True
        except (base64.binascii.Error, ValueError):
            return False

    def __is_auth_cert_data(self, pem_str):
        """
        Valida si una cadena tiene formato PEM válido.

        El formato PEM debe cumplir con los siguientes criterios:
            - Comienza con una línea "-----BEGIN [LABEL]-----"
            - Termina con una línea "-----END [LABEL]-----"
            - Contiene contenido Base64 válido entre las líneas BEGIN y END

        **Ejemplos de PEM Válidos:**
            ```
            -----BEGIN CERTIFICATE-----
            MIIDdzCCAl+gAwIBAgIEbGzVnzANBgkqhkiG9w0BAQsFADBvMQswCQYDVQQGEwJV
            ...
            -----END CERTIFICATE-----
            ```

        **Ejemplos de PEM Inválidos:**
            - Falta la línea de inicio o fin.
            - Contenido no codificado en Base64.
            - Etiquetas de BEGIN y END que no coinciden.

        :param str pem_str: La cadena a validar.
        :return: True si la cadena tiene formato PEM válido, False en caso contrario.
        :rtype: bool
        """
        if pem_str is None:
            return False
        # Expresión regular para validar el formato PEM
        patron = re.compile(
            r'-----BEGIN ([A-Z ]+)-----\s+'
            r'([A-Za-z0-9+/=\s]+)'
            r'-----END \1-----$',
            re.MULTILINE
        )

        # Intentar hacer match con el patrón
        match = patron.fullmatch(pem_str.strip())
        if not match:
            return False

        # Extraer el contenido Base64
        base64_content = match.group(2).replace('\n', '').replace('\r', '').strip()

        # Verificar que el contenido Base64 sea válido
        try:
            base64.b64decode(base64_content, validate=True)
            return True
        except (base64.binascii.Error, ValueError):
            return False

    def _get_auth_pass(self):
        '''
        Obtiene la autenticación de tipo 'pass'.

        :return: Información de autenticación.
        :rtype: dict
        :raises ApiException: Si falta información de autenticación.
        '''
        if 'pass' in self.auth:
            if 'rut' not in self.auth['pass']:
                raise ApiException('auth.pass.rut missing.')
            if self.auth['pass']['rut'] == '' or self.auth['pass']['rut'] is None:
                raise ApiException('auth.pass.rut empty.')
            if 'clave' not in self.auth['pass']:
                raise ApiException('auth.pass.clave missing.')
            if self.auth['pass']['clave'] == '' or self.auth['pass']['clave'] is None:
                raise ApiException('auth.pass.clave empty.')
        elif 'cert' in self.auth:
            if 'cert-data' in self.auth['cert'] and self.auth['cert']['cert-data'] == '' and self.auth['cert']['cert-data'] is None:
                raise ApiException('auth.cert.cert-data empty.')
            if 'pkey-data' in self.auth['cert'] and self.auth['cert']['pkey-data'] == '' and self.auth['cert']['pkey-data'] is None:
                raise ApiException('auth.cert.pkey-data empty.')
            if 'file-data' in self.auth['cert'] and self.auth['cert']['file-data'] == '' and self.auth['cert']['file-data'] is None:
                raise ApiException('auth.cert.file-data empty.')
            if 'file-pass' in self.auth['cert'] and self.auth['cert']['file-pass'] == '' and self.auth['cert']['file-pass'] is None:
                raise ApiException('auth.cert.file-pass empty.')
        else:
            raise ApiException('auth.pass or auth.cert missing.')
        return self.auth



#
# API Gateway: Cliente de API en Python.
# Copyright (C) API Gateway <https://www.apigateway.cl>
#
# Este programa es software libre: usted puede redistribuirlo y/o modificarlo
# bajo los términos de la GNU Lesser General Public License (LGPL) publicada
# por la Fundación para el Software Libre, ya sea la versión 3 de la Licencia,
# o (a su elección) cualquier versión posterior de la misma.
#
# Este programa se distribuye con la esperanza de que sea útil, pero SIN
# GARANTÍA ALGUNA; ni siquiera la garantía implícita MERCANTIL o de APTITUD
# PARA UN PROPÓSITO DETERMINADO. Consulte los detalles de la GNU Lesser General
# Public License (LGPL) para obtener una información más detallada.
#
# Debería haber recibido una copia de la GNU Lesser General Public License
# (LGPL) junto a este programa. En caso contrario, consulte
# <http://www.gnu.org/licenses/lgpl.html>.
#

'''
Módulo para obtener datos de los contribuyentes a través del SII.

Para más información sobre la API, consulte la `documentación completa de los
Contribuyentes <https://developers.apigateway.cl/#c88f90b6-36bb-4dc2-ba93-6e418ff42098>`_.
'''

from .. import ApiBase

class Contribuyentes(ApiBase):
    '''
    Cliente específico para interactuar con los endpoints de contribuyentes
    de la API de API Gateway.

    Hereda de ApiBase y utiliza su funcionalidad para realizar solicitudes a la API.
    '''

    def situacion_tributaria(self, rut):
        '''
        Obtiene la situación tributaria de un contribuyente.

        :param str rut: RUT del contribuyente.
        :return: Respuesta JSON con la situación tributaria del contribuyente.
        :rtype: dict
        '''
        url = '/sii/contribuyentes/situacion_tributaria/tercero/%(rut)s' % {
            'rut': rut
        }
        response = self.client.get(url)
        return response.json()

    def verificar_rut(self, serie):
        '''
        Verifica el RUT de un contribuyente.

        :param str serie: Serie del RUT a verificar.
        :return: Respuesta JSON con la verificación del RUT.
        :rtype: dict
        '''
        url = '/sii/contribuyentes/rut/verificar/%(serie)s' % {
            'serie': serie
        }
        response = self.client.get(url)
        return response.json()



#
# API Gateway: Cliente de API en Python.
# Copyright (C) API Gateway <https://www.apigateway.cl>
#
# Este programa es software libre: usted puede redistribuirlo y/o modificarlo
# bajo los términos de la GNU Lesser General Public License (LGPL) publicada
# por la Fundación para el Software Libre, ya sea la versión 3 de la Licencia,
# o (a su elección) cualquier versión posterior de la misma.
#
# Este programa se distribuye con la esperanza de que sea útil, pero SIN
# GARANTÍA ALGUNA; ni siquiera la garantía implícita MERCANTIL o de APTITUD
# PARA UN PROPÓSITO DETERMINADO. Consulte los detalles de la GNU Lesser General
# Public License (LGPL) para obtener una información más detallada.
#
# Debería haber recibido una copia de la GNU Lesser General Public License
# (LGPL) junto a este programa. En caso contrario, consulte
# <http://www.gnu.org/licenses/lgpl.html>.
#

'''
Módulo para obtener indicadores desde el SII.

Para más información sobre la API, consulte la `documentación completa de los
Indicadores <https://developers.apigateway.cl/#65aa568c-4c5a-448b-9f3b-95c3d9153e4d>`_.
'''

from .. import ApiBase

class Uf(ApiBase):
    '''
    Cliente específico para interactuar con los endpoints de valores de UF
    (Unidad de Fomento) de la API de API Gateway.

    Provee métodos para obtener valores de UF anuales, mensuales y diarios.
    '''

    def anual(self, anio):
        '''
        Obtiene los valores de la UF para un año específico.

        :param int anio: Año para el cual se quieren obtener los valores de la UF.
        :return: Respuesta JSON con los valores de la UF del año especificado.
        :rtype: dict
        '''
        anio = str(anio)
        url = '/sii/indicadores/uf/anual/%(anio)s' % {'anio': anio}
        response = self.client.get(url)
        datos = response.json()
        return datos[anio] if anio in datos else {}

    def mensual(self, periodo):
        '''
        Obtiene los valores de la UF para un mes específico.

        :param str periodo: Período en formato AAAAMM (año y mes).
        :return: Respuesta JSON con los valores de la UF del mes especificado.
        :rtype: dict
        '''
        anio, mes = periodo[:4], periodo[4:6]
        url = '/sii/indicadores/uf/anual/%(anio)s/%(mes)s' % {
            'anio': anio, 'mes': mes
        }
        response = self.client.get(url)
        datos = response.json()
        return datos[periodo] if periodo in datos else {}

    def diario(self, dia):
        '''
        Obtiene el valor de la UF para un día específico.

        :param str dia: Fecha en formato AAAA-MM-DD.
        :return: Valor de la UF para el día especificado.
        :rtype: float
        '''
        anio, mes, dia = dia.split('-')
        url = '/sii/indicadores/uf/anual/%(anio)s/%(mes)s/%(dia)s' % {
            'anio': anio, 'mes': mes, 'dia': dia
        }
        response = self.client.get(url)
        datos = response.json()
        key = '%(anio)s%(mes)s%(dia)s' % {
            'anio': anio, 'mes': mes, 'dia': dia
        }
        return float(datos[key]) if key in datos else 0



#
# API Gateway: Cliente de API en Python.
# Copyright (C) API Gateway <https://www.apigateway.cl>
#
# Este programa es software libre: usted puede redistribuirlo y/o modificarlo
# bajo los términos de la GNU Lesser General Public License (LGPL) publicada
# por la Fundación para el Software Libre, ya sea la versión 3 de la Licencia,
# o (a su elección) cualquier versión posterior de la misma.
#
# Este programa se distribuye con la esperanza de que sea útil, pero SIN
# GARANTÍA ALGUNA; ni siquiera la garantía implícita MERCANTIL o de APTITUD
# PARA UN PROPÓSITO DETERMINADO. Consulte los detalles de la GNU Lesser General
# Public License (LGPL) para obtener una información más detallada.
#
# Debería haber recibido una copia de la GNU Lesser General Public License
# (LGPL) junto a este programa. En caso contrario, consulte
# <http://www.gnu.org/licenses/lgpl.html>.
#

'''
Módulo para interactuar con la sección MiSii de un contribuyente en el sitio web del SII.

Para más información sobre la API, consulte la `documentación completa de
MiSii <https://developers.apigateway.cl/#b585f374-f106-46a9-9f47-666d478b8ac8>`_.
'''

from .. import ApiBase

class Contribuyente(ApiBase):
    '''
    Cliente específico para interactuar con los endpoints de un Contribuyente
    de MiSii de la API de API Gateway.

    Hereda de ApiBase y utiliza su funcionalidad para realizar solicitudes a la API.
    '''

    def __init__(self, identificador, clave, **kwargs):
        super().__init__(identificador = identificador, clave = clave, **kwargs)

    def datos(self, auth_cache = None):
        '''
        Obtiene los datos de MiSii del contribuyente autenticado en el SII.

        :return: Respuesta JSON con los datos del contribuyente.
        :rtype: dict
        '''
        url = '/sii/misii/contribuyente/datos'
        body = {
            'auth': self._get_auth_pass()
        }
        if auth_cache is not None:
            url += '&auth_cache=0'
        response = self.client.post(url, data = body)
        return response.json()



#
# API Gateway: Cliente de API en Python.
# Copyright (C) API Gateway <https://www.apigateway.cl>
#
# Este programa es software libre: usted puede redistribuirlo y/o modificarlo
# bajo los términos de la GNU Lesser General Public License (LGPL) publicada
# por la Fundación para el Software Libre, ya sea la versión 3 de la Licencia,
# o (a su elección) cualquier versión posterior de la misma.
#
# Este programa se distribuye con la esperanza de que sea útil, pero SIN
# GARANTÍA ALGUNA; ni siquiera la garantía implícita MERCANTIL o de APTITUD
# PARA UN PROPÓSITO DETERMINADO. Consulte los detalles de la GNU Lesser General
# Public License (LGPL) para obtener una información más detallada.
#
# Debería haber recibido una copia de la GNU Lesser General Public License
# (LGPL) junto a este programa. En caso contrario, consulte
# <http://www.gnu.org/licenses/lgpl.html>.
#

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(

    name='apigatewaycl',

    # Versions should comply with PEP440
    version='5.2.0',

    description='Cliente para realizar la integración con los servicios web de API Gateway desde Python.',
    long_description="\n"+long_description,

    # The project's main homepage.
    url='https://github.com/apigatewaycl/apigateway-api-client-python',

    # Author details
    author='API Gateway',
    author_email='dev@apigateway.cl',

    # Choose your license
    license='LGPL',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',
        #'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
    ],

    # What does your project relate to?
    keywords='sii chile',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # List run-time dependencies here
    install_requires=['requests'],

)
